import datetime as dt
import re
from urllib.error import URLError

import pandas as pd
import httpx
from pandas.tseries.offsets import MonthEnd

from econuy import load_dataset
from econuy.base import Dataset, DatasetMetadata
from econuy.utils.extras import FISCAL_SHEETS, taxes_columns
from econuy.utils.operations import get_name_from_function, get_download_sources
from econuy.utils.retrieval import get_with_ssl_context


def _get_fiscal_balances(dataset_name: str) -> Dataset:
    """Helper function. See any of the `fiscal_balance_...()` functions."""
    sources = get_download_sources("fiscal_balances")
    response = httpx.get(sources["main"])
    url = re.findall(r"(http\S+Resultados.+\.xlsx)'", response.text)[0]
    xls = pd.ExcelFile(url)
    output = {}
    dataset_details = FISCAL_SHEETS[dataset_name]
    output = (
        pd.read_excel(xls, sheet_name=dataset_details["sheet"])
        .dropna(axis=0, thresh=4)
        .dropna(axis=1, thresh=4)
        .transpose()
        .set_index(2, drop=True)
    )
    output = output[output.index.notnull()].rename_axis(None)
    output.index = output.index + MonthEnd(1)
    output = output.apply(pd.to_numeric, errors="coerce")
    output = output.rename_axis(None)

    ids = [f"{dataset_name}_{i}" for i in range(output.shape[1])]
    output.columns = ids
    spanish_names = dataset_details["colnames"]
    spanish_names = [{"es": x} for x in spanish_names]

    base_metadata = {
        "area": "Public sector",
        "currency": "UYU",
        "inflation_adjustment": None,
        "unit": "Millions",
        "seasonal_adjustment": None,
        "frequency": "ME",
        "time_series_type": "Flow",
        "cumulative_periods": 1,
        "transformations": [],
    }
    metadata = DatasetMetadata.from_cast(
        dataset_name, base_metadata, output.columns, spanish_names
    )
    dataset = Dataset(dataset_name, output, metadata)

    return dataset


def fiscal_balance_global_public_sector() -> Dataset:
    """Get fiscal balance data for the consolidated public sector.

    Returns
    -------
    Monthly fiscal balance for the consolidated public sector : Dataset

    """
    name = get_name_from_function()
    return _get_fiscal_balances(name)


def fiscal_balance_nonfinancial_public_sector() -> Dataset:
    """Get fiscal balance data for the non-financial public sector.

    Returns
    -------
    Monthly fiscal balance for the non-financial public sector : Dataset

    """
    name = get_name_from_function()
    return _get_fiscal_balances(name)


def fiscal_balance_central_government() -> Dataset:
    """Get fiscal balance data for the central government + BPS.

    Returns
    -------
    Monthly fiscal balance for the central government + BPS : Dataset

    """
    name = get_name_from_function()
    return _get_fiscal_balances(name)


def fiscal_balance_soe() -> Dataset:
    """Get fiscal balance data for public enterprises.

    Returns
    -------
    Monthly fiscal balance for public enterprises : Dataset

    """
    name = get_name_from_function()
    return _get_fiscal_balances(name)


def fiscal_balance_ancap() -> Dataset:
    """Get fiscal balance data for ANCAP.

    Returns
    -------
    Monthly fiscal balance for ANCAP : Dataset

    """
    name = get_name_from_function()
    return _get_fiscal_balances(name)


def fiscal_balance_ute() -> Dataset:
    """Get fiscal balance data for UTE.

    Returns
    -------
    Monthly fiscal balance for UTE : Dataset

    """
    name = get_name_from_function()
    return _get_fiscal_balances(name)


def fiscal_balance_antel() -> Dataset:
    """Get fiscal balance data for ANTEL.

    Returns
    -------
    Monthly fiscal balance for ANTEL : Dataset

    """
    name = get_name_from_function()
    return _get_fiscal_balances(name)


def fiscal_balance_ose() -> Dataset:
    """Get fiscal balance data for OSE.

    Returns
    -------
    Monthly fiscal balance for OSE : Dataset

    """
    name = get_name_from_function()
    return _get_fiscal_balances(name)


def tax_revenue() -> Dataset:
    """
    Get tax revenues data.

    This retrieval function requires that Ghostscript and Tkinter be found in
    your system.

    Returns
    -------
    Monthly tax revenues : Dataset

    """
    name = get_name_from_function()
    sources = get_download_sources(name)
    r = httpx.get(sources["main"], timeout=20)
    url = re.findall(
        "https://[A-z0-9-/\.]+Recaudaci%C3%B3n%20por%20impuesto%20-%20Series%20mensuales.csv",
        r.text,
    )[0]
    historical = pd.read_csv(
        url, skiprows=2, encoding="Latin", sep=";", thousands="."
    ).iloc[:, 3:41]
    historical.index = pd.date_range("1982-01-31", periods=len(historical), freq="ME")
    historical.columns = taxes_columns
    historical = historical.div(1000000)

    try:
        aux = historical.copy()
        latest = pd.read_csv(sources["pdfs"], index_col=0, parse_dates=True)
        latest.columns = [
            "IVA - Valor Agregado",
            "IMESI - Específico Interno",
            "IMEBA - Enajenación de Bienes Agropecuarios",
            "IRAE - Rentas de Actividades Económicas",
            "IRPF Cat I - Renta de las Personas Físicas",
            "IRPF Cat II - Rentas de las Personas Físicas",
            "IASS - Asistencia a la Seguridad Social",
            "IRNR - Rentas de No Residentes",
            "Impuesto de Educación Primaria",
            "Recaudación Total de la DGI",
        ]
        latest = latest.loc[[x not in aux.index for x in latest.index]]
        for col in latest.columns:
            for date in latest.index:
                prev_year = date + MonthEnd(-12)
                aux.loc[date, col] = aux.loc[prev_year, col] * (
                    1 + latest.loc[date, col] / 100
                )
        output = pd.concat([aux, latest], sort=False)
        output = output.loc[~output.index.duplicated(keep="first")]
    except Exception as e:
        print(f"Could not get PDF data | {e}")
        output = historical.copy()

    output = output.apply(pd.to_numeric, errors="coerce")
    output = output.rename_axis(None)

    spanish_names = output.columns
    spanish_names = [{"es": x} for x in spanish_names]
    ids = [f"{name}_{i}" for i in range(output.shape[1])]
    output.columns = ids

    base_metadata = {
        "area": "Public sector",
        "currency": "UYU",
        "inflation_adjustment": None,
        "unit": "Millions",
        "seasonal_adjustment": None,
        "frequency": "ME",
        "time_series_type": "Flow",
        "cumulative_periods": 1,
        "transformations": [],
    }
    metadata = DatasetMetadata.from_cast(
        name, base_metadata, output.columns, spanish_names
    )
    dataset = Dataset(name, output, metadata)

    return dataset


def _get_public_debt(dataset_name: str) -> Dataset:
    """Helper function. See any of the `public_debt_...()` functions."""
    sources = get_download_sources("public_debt")
    colnames = [
        "Total deuda",
        "Plazo contractual: hasta 1 año",
        "Plazo contractual: entre 1 y 5 años",
        "Plazo contractual: más de 5 años",
        "Plazo residual: hasta 1 año",
        "Plazo residual: entre 1 y 5 años",
        "Plazo residual: más de 5 años",
        "Moneda: pesos",
        "Moneda: dólares",
        "Moneda: euros",
        "Moneda: yenes",
        "Moneda: DEG",
        "Moneda: otras",
        "Residencia: no residentes",
        "Residencia: residentes",
    ]
    try:
        xls = pd.ExcelFile(sources["main"])
    except URLError as err:
        if "SSL: CERTIFICATE_VERIFY_FAILED" in str(err):
            r_bytes = get_with_ssl_context("bcu", sources["main"])
            xls = pd.ExcelFile(r_bytes)

    if dataset_name == "public_debt_global_public_sector":
        gps_raw = pd.read_excel(
            xls,
            sheet_name="SPG2",
            usecols="B:Q",
            index_col=0,
            skiprows=10,
            nrows=(dt.datetime.now().year - 1999) * 4,
        )
        output = gps_raw.dropna(thresh=2)
        output.index = pd.date_range(
            start="1999-12-31", periods=len(output), freq="QE-DEC"
        )
        output.columns = colnames

    elif dataset_name == "public_debt_nonfinancial_public_sector":
        nfps_raw = pd.read_excel(
            xls, sheet_name="SPNM bruta", usecols="B:O", index_col=0
        )
        loc = nfps_raw.index.get_loc(
            "9. Deuda Bruta del Sector Público no " "monetario por plazo y  moneda."
        )
        output = nfps_raw.iloc[loc + 5 :, :].dropna(how="any")
        output.index = pd.date_range(
            start="1999-12-31", periods=len(output), freq="QE-DEC"
        )
        nfps_extra_raw = pd.read_excel(
            xls,
            sheet_name="SPNM bruta",
            usecols="O:P",
            skiprows=11,
            nrows=(dt.datetime.now().year - 1999) * 4,
        )
        nfps_extra = nfps_extra_raw.dropna(how="all")
        nfps_extra.index = output.index
        output = pd.concat([output, nfps_extra], axis=1)
        output.columns = colnames

    elif dataset_name == "public_debt_central_bank":
        cb_raw = pd.read_excel(
            xls,
            sheet_name="BCU bruta",
            usecols="B:O",
            index_col=0,
            skiprows=(dt.datetime.now().year - 1999) * 8 + 20,
        )
        output = cb_raw.dropna(how="any")
        output.index = pd.date_range(
            start="1999-12-31", periods=len(output), freq="QE-DEC"
        )
        cb_extra_raw = pd.read_excel(
            xls,
            sheet_name="BCU bruta",
            usecols="O:P",
            skiprows=11,
            nrows=(dt.datetime.now().year - 1999) * 4,
        )
        bcu_extra = cb_extra_raw.dropna(how="all")
        bcu_extra.index = output.index
        output = pd.concat([output, bcu_extra], axis=1)
        output.columns = colnames

    else:
        assets_raw = pd.read_excel(
            xls,
            sheet_name="Activos Neta",
            usecols="B,C,D,K",
            index_col=0,
            skiprows=13,
            nrows=(dt.datetime.now().year - 1999) * 4 - 1,
        )
        output = assets_raw.dropna(how="any")
        output.index = pd.date_range(
            start="1999-12-31", periods=len(output), freq="QE-DEC"
        )
        output.columns = ["Total activos", "Sector público no monetario", "BCU"]

    output = output.apply(pd.to_numeric, errors="coerce")

    spanish_names = output.columns
    spanish_names = [{"es": x} for x in spanish_names]
    ids = [f"{dataset_name}_{i}" for i in range(output.shape[1])]
    output.columns = ids

    base_metadata = {
        "area": "Public sector",
        "currency": "USD",
        "inflation_adjustment": None,
        "unit": "Millions",
        "seasonal_adjustment": None,
        "frequency": "ME",
        "time_series_type": "Stock",
        "cumulative_periods": 1,
        "transformations": [],
    }
    metadata = DatasetMetadata.from_cast(
        dataset_name, base_metadata, output.columns, spanish_names
    )
    dataset = Dataset(dataset_name, output, metadata)

    return dataset


def public_debt_global_public_sector() -> Dataset:
    """Get public debt data for the consolidated public sector.

    Returns
    -------
    Quarterly public debt data for the consolidated public sector: Dataset

    """
    name = get_name_from_function()
    return _get_public_debt(name)


def public_debt_nonfinancial_public_sector() -> Dataset:
    """Get public debt data for the non-financial public sector.

    Returns
    -------
    Quarterly public debt data for the non-financial public sector: Dataset

    """
    name = get_name_from_function()
    return _get_public_debt(name)


def public_debt_central_bank() -> Dataset:
    """Get public debt data for the central bank

    Returns
    -------
    Quarterly public debt data for the central bank : Dataset

    """
    name = get_name_from_function()
    return _get_public_debt(name)


def public_assets() -> Dataset:
    """Get public sector assets data.

    Returns
    -------
    Quarterly public sector assets: Dataset

    """
    name = get_name_from_function()
    return _get_public_debt(name)


def net_public_debt_global_public_sector(*args, **kwargs) -> Dataset:
    """
    Get net public debt excluding deposits at the central bank.


    Returns
    -------
    Net public debt excl. deposits at the central bank : Dataset

    """
    name = get_name_from_function()

    gross_debt = load_dataset(
        "public_debt_global_public_sector", *args, **kwargs
    ).to_named()[["Total deuda"]]
    assets = load_dataset("public_assets", *args, **kwargs).to_named()[
        ["Total activos"]
    ]
    gross_debt.columns = ["Deuda neta del sector público global excl. encajes"]
    assets.columns = gross_debt.columns
    deposits = (
        load_dataset("international_reserves", *args, **kwargs)
        .resample("QE-DEC", "last")
        .to_named()[["Obligaciones en ME con el sector financiero"]]
    )
    deposits = deposits.reindex(gross_debt.index).squeeze()
    output = gross_debt.add(assets).add(deposits, axis=0).dropna()
    output = output.rename_axis(None)

    spanish_names = output.columns
    spanish_names = [{"es": x} for x in spanish_names]
    ids = [f"{name}_{i}" for i in range(output.shape[1])]
    output.columns = ids

    base_metadata = {
        "area": "Public sector",
        "currency": "USD",
        "inflation_adjustment": None,
        "unit": "Millions",
        "seasonal_adjustment": None,
        "frequency": "QE-DEC",
        "time_series_type": "Stock",
        "cumulative_periods": 1,
        "transformations": [],
    }
    metadata = DatasetMetadata.from_cast(
        name, base_metadata, output.columns, spanish_names
    )
    dataset = Dataset(name, output, metadata)

    return dataset


def fiscal_balance_summary(*args, **kwargs) -> Dataset:
    """
    Get the summary fiscal balance table found in the `Budget Law
    <https://www.gub.uy/contaduria-general-nacion/sites/
    contaduria-general-nacion/files/2020-09/
    Mensaje%20y%20Exposici%C3%B3n%20de%20motivos.pdf>`_. Includes adjustments
    for the `Social Security Fund <https://www.impo.com.uy/bases/decretos/
    71-2018/25>`_.

    Returns
    -------
    Summary fiscal balance table : Dataset

    """
    name = get_name_from_function()

    datasets = {}
    for dataset_name in [
        "fiscal_balance_global_public_sector",
        "fiscal_balance_nonfinancial_public_sector",
        "fiscal_balance_central_government",
        "fiscal_balance_soe",
    ]:
        d = load_dataset(dataset_name, *args, **kwargs).to_detailed()
        d.columns = d.columns.get_level_values(0)
        datasets.update({dataset_name: d})
    gps = datasets["fiscal_balance_global_public_sector"]
    nfps = datasets["fiscal_balance_nonfinancial_public_sector"]
    gc = datasets["fiscal_balance_central_government"]
    pe = datasets["fiscal_balance_soe"]
    proc = pd.DataFrame(index=gps.index)

    proc["Ingresos: GC-BPS"] = gc["Ingresos: GC-BPS"]
    proc["Ingresos: GC-BPS ex. FSS"] = (
        gc["Ingresos: GC-BPS"] - gc["Ingresos: FSS - Cincuentones"]
    )
    proc["Ingresos: GC"] = gc["Ingresos: GC"]
    proc["Ingresos: DGI"] = gc["Ingresos: DGI"]
    proc["Ingresos: Comercio ext."] = gc["Ingresos: Comercio ext."]
    proc["Ingresos: Otros"] = (
        gc["Ingresos: GC"] - gc["Ingresos: DGI"] - gc["Ingresos: Comercio ext."]
    )
    proc["Ingresos: BPS"] = gc["Ingresos: BPS neto"]
    proc["Ingresos: FSS - Cincuentones"] = gc["Ingresos: FSS - Cincuentones"]
    proc["Ingresos: BPS ex FSS"] = (
        gc["Ingresos: BPS neto"] - gc["Ingresos: FSS - Cincuentones"]
    )
    proc["Egresos: Primarios GC-BPS"] = gc["Egresos: GC-BPS"] - gc["Intereses: Total"]
    proc["Egresos: Primarios corrientes GC-BPS"] = (
        proc["Egresos: Primarios GC-BPS"] - gc["Egresos: Inversión"].squeeze()
    )
    proc["Egresos: Remuneraciones"] = gc["Egresos: Remuneraciones"]
    proc["Egresos: No personales"] = gc["Egresos: No personales"]
    proc["Egresos: Pasividades"] = gc["Egresos: Pasividades"]
    proc["Egresos: Transferencias"] = gc["Egresos: Transferencias"]
    proc["Egresos: Inversión"] = gc["Egresos: Inversión"]
    proc["Resultado: Primario GC-BPS"] = (
        proc["Ingresos: GC-BPS"] - proc["Egresos: Primarios GC-BPS"]
    )
    proc["Resultado: Primario GC-BPS ex FSS"] = (
        proc["Ingresos: GC-BPS ex. FSS"] - proc["Egresos: Primarios GC-BPS"]
    )
    proc["Intereses: GC-BPS"] = gc["Intereses: Total"]
    proc["Intereses: FSS - Cincuentones"] = gc["Intereses: FSS - Cincuentones"]
    proc["Intereses: GC-BPS ex FSS"] = (
        proc["Intereses: GC-BPS"] - proc["Intereses: FSS - Cincuentones"]
    )
    proc["Resultado: Global GC-BPS"] = (
        proc["Resultado: Primario GC-BPS"] - proc["Intereses: GC-BPS"]
    )
    proc["Resultado: Global GC-BPS ex FSS"] = (
        proc["Resultado: Primario GC-BPS ex FSS"] - proc["Intereses: GC-BPS ex FSS"]
    )

    proc["Resultado: Primario corriente EEPP"] = nfps[
        "Ingresos: Res. primario corriente EEPP"
    ]
    proc["Egresos: Inversiones EEPP"] = pe["Egresos: Inversiones"]
    proc["Resultado: Primario EEPP"] = (
        proc["Resultado: Primario corriente EEPP"] - proc["Egresos: Inversiones EEPP"]
    )
    proc["Intereses: EEPP"] = pe["Intereses"]
    proc["Resultado: Global EEPP"] = (
        proc["Resultado: Primario EEPP"] - proc["Intereses: EEPP"]
    )

    proc["Resultado: Primario intendencias"] = nfps["Resultado: Primario intendencias"]
    proc["Intereses: Intendencias"] = nfps["Intereses: Intendencias"]
    proc["Resultado: Global intendencias"] = (
        proc["Resultado: Primario intendencias"] - proc["Intereses: Intendencias"]
    )

    proc["Resultado: Primario BSE"] = nfps["Resultado: Primario BSE"]
    proc["Intereses: BSE"] = nfps["Intereses: BSE"]
    proc["Resultado: Global BSE"] = (
        proc["Resultado: Primario BSE"] - proc["Intereses: BSE"]
    )

    proc["Resultado: Primario resto SPNF"] = (
        proc["Resultado: Primario EEPP"]
        + proc["Resultado: Primario intendencias"]
        + proc["Resultado: Primario BSE"]
    )
    proc["Intereses: Resto SPNF"] = (
        proc["Intereses: EEPP"]
        + proc["Intereses: Intendencias"]
        + proc["Intereses: BSE"]
    )
    proc["Resultado: Global resto SPNF"] = (
        proc["Resultado: Global EEPP"]
        + proc["Resultado: Global intendencias"]
        + proc["Resultado: Global BSE"]
    )
    proc["Resultado: Primario SPNF"] = nfps["Resultado: Primario SPNF"]
    proc["Resultado: Primario SPNF ex FSS"] = (
        proc["Resultado: Primario SPNF"] - proc["Ingresos: FSS - Cincuentones"]
    )
    proc["Intereses: SPNF"] = nfps["Intereses: Totales"]
    proc["Intereses: SPNF ex FSS"] = (
        proc["Intereses: SPNF"] - proc["Intereses: FSS - Cincuentones"]
    )
    proc["Resultado: Global SPNF"] = nfps["Resultado: Global SPNF"]
    proc["Resultado: Global SPNF ex FSS"] = (
        proc["Resultado: Primario SPNF ex FSS"] - proc["Intereses: SPNF ex FSS"]
    )

    proc["Resultado: Primario BCU"] = gps["Resultado: Primario BCU"]
    proc["Intereses: BCU"] = gps["Intereses: BCU"]
    proc["Resultado: Global BCU"] = gps["Resultado: Global BCU"]

    proc["Resultado: Primario SPC"] = gps["Resultado: Primario SPC"]
    proc["Resultado: Primario SPC ex FSS"] = (
        proc["Resultado: Primario SPNF ex FSS"] + proc["Resultado: Primario BCU"]
    )
    proc["Intereses: SPC"] = proc["Intereses: SPNF"] + proc["Intereses: BCU"]
    proc["Intereses: SPC ex FSS"] = (
        proc["Intereses: SPNF ex FSS"] + proc["Intereses: BCU"]
    )
    proc["Resultado: Global SPC"] = (
        proc["Resultado: Global SPNF"] + proc["Resultado: Global BCU"]
    )
    proc["Resultado: Global SPC ex FSS"] = (
        proc["Resultado: Global SPNF ex FSS"] + proc["Resultado: Global BCU"]
    )
    output = proc
    output = output.rename_axis(None)
    spanish_names = output.columns
    spanish_names = [{"es": x} for x in spanish_names]

    ids = [f"{name}_{i}" for i in range(output.shape[1])]
    output.columns = ids

    base_metadata = {
        "area": "Public sector",
        "currency": "UYU",
        "inflation_adjustment": None,
        "unit": "Millions",
        "seasonal_adjustment": None,
        "frequency": "ME",
        "time_series_type": "Flow",
        "cumulative_periods": 1,
        "transformations": [],
    }
    metadata = DatasetMetadata.from_cast(
        name, base_metadata, output.columns, spanish_names
    )
    dataset = Dataset(name, output, metadata)

    return dataset



import importlib
import datetime as dt
import inspect
import os
from typing import Union, List, Optional, Dict, Literal
from pathlib import Path
from urllib.error import URLError
from json.decoder import JSONDecodeError
from concurrent import futures

import pandas as pd
from httpx import ReadTimeout
from opnieuw import retry
from tqdm.auto import tqdm

from econuy.utils.operations import REGISTRY, read_dataset, get_data_dir
from econuy.base import Dataset


OUTDATED_DELTA_THRESHOLD = dt.timedelta(days=1)  # TODO: Use an env var or config file


@retry(
    retry_on_exceptions=(ConnectionError, URLError, JSONDecodeError, ReadTimeout),
    max_calls_total=4,
    retry_window_after_first_call_in_seconds=30,
)
def load_dataset(
    name: str,
    data_dir: Union[str, Path, None] = None,
    skip_cache: bool = False,
    force_overwrite: bool = False,
    skip_update: bool = False,
) -> Dataset:
    """
    Load a dataset by name, optionally skipping cache and forcing overwrite.

    Parameters
    ----------
    name : str
        The name of the dataset to load.
    data_dir : Union[str, Path, None], optional
        The directory where the dataset is stored or will be stored. If None,
        the default data directory is used. Default is None.
    skip_cache : bool, optional
        If True, the cache will be skipped and a new dataset will be retrieved.
        Default is False.
    force_overwrite : bool, optional
        If True, the existing dataset will be overwritten. Default is False.
    skip_update : bool, optional
        If True, the dataset will not be updated if it already exists. Default is False.

    Returns
    -------
    Dataset
        The loaded dataset.

    Raises
    ------
    ValueError
        If the dataset name is not available in the registry.
    AssertionError
        If the existing dataset has changed and force_overwrite is False.
    """
    data_dir = data_dir or get_data_dir()
    data_dir = Path(data_dir)
    data_dir.mkdir(parents=True, exist_ok=True, mode=0o755)

    if not skip_cache:
        existing_dataset = read_dataset(name, data_dir)
        if existing_dataset is not None:
            created_at = existing_dataset.metadata.created_at
            if (
                dt.datetime.now() - created_at
            ) < OUTDATED_DELTA_THRESHOLD or skip_update:
                return existing_dataset
            else:
                print(
                    f"Dataset {name} exists in cache but is outdated "
                    f"(created at {created_at.strftime('%Y-%m-%d %H:%M:%S')}). "
                    "Retrieving new data."
                )

    try:
        dataset_metadata = REGISTRY[name]
    except KeyError:
        raise ValueError(f"Dataset {name} not available.")

    function_string = dataset_metadata["function"]
    module, function = function_string.split(".")
    path_prefix = "econuy.retrieval."
    module = importlib.import_module(path_prefix + module)
    dataset_retriever = getattr(module, function)

    signature = inspect.signature(dataset_retriever)
    parameters = signature.parameters
    if parameters:
        dataset = dataset_retriever(data_dir, skip_cache, force_overwrite, skip_update)
    else:
        dataset = dataset_retriever()

    if not force_overwrite:
        existing_dataset = read_dataset(name, data_dir)
        if existing_dataset is not None:
            try:
                check_updated_dataset(existing_dataset, dataset)
                dataset.save(data_dir)
            except AssertionError as exc:
                print(f"Dataset {name} has changed. Will not overwrite. Error: {exc}")
        else:
            dataset.save(data_dir)
    else:
        dataset.save(data_dir)

    return dataset


def load_datasets_parallel(
    names: List[str],
    data_dir: Union[str, Path, None] = None,
    skip_cache: bool = False,
    force_overwrite: bool = False,
    skip_update: bool = False,
    max_workers: Optional[int] = None,
    executor_type: Literal["thread", "process"] = "thread",
) -> Dict[str, Dataset]:
    """
    Load multiple datasets in parallel using either threading or multiprocessing.

    Parameters
    ----------
    names : List[str]
        List of dataset names to load.
    data_dir : Union[str, Path, None], optional
        Directory where datasets are stored. If None, a default directory is used.
    skip_cache : bool, optional
        If True, skip loading from cache. Default is False.
    force_overwrite : bool, optional
        If True, force overwrite existing datasets. Default is False.
    skip_update : bool, optional
        If True, skip updating datasets that already exist. Default is False.
    max_workers : Optional[int], optional
        Maximum number of workers to use for parallel loading. If None, it will use the default number of workers.
    executor_type : Literal["thread", "process"], optional
        Type of executor to use for parallel loading. Can be "thread" for ThreadPoolExecutor or "process" for ProcessPoolExecutor. Default is "thread".

    Returns
    -------
    Dict[str, Dataset]
        A dictionary where keys are dataset names and values are the loaded datasets.

    Raises
    ------
    Exception
        If there is an error loading any of the datasets, it will be printed and the dataset will be skipped.
    """
    datasets = {}

    # We first pick an executor, then get the default workers used in the stdlib code.
    # If max_workers is not set, we use the default number of workers.
    # In both cases we limit the number of workers to the number of datasets.
    if executor_type == "thread":
        executor_class = futures.ThreadPoolExecutor
        default_workers = min(32, (os.cpu_count() or 1) + 4)
    elif executor_type == "process":
        executor_class = futures.ProcessPoolExecutor
        default_workers = os.cpu_count() or 1

    workers = max_workers or default_workers
    workers = min(workers, len(names))

    with executor_class(workers) as executor:
        future_to_name = {
            executor.submit(
                load_dataset, name, data_dir, skip_cache, force_overwrite, skip_update
            ): name
            for name in names
        }
        with tqdm(total=len(names), desc="Loading datasets") as pbar:
            for future in futures.as_completed(future_to_name):
                name = future_to_name[future]
                pbar.set_postfix_str(name)
                try:
                    dataset = future.result()
                    datasets[name] = dataset
                except Exception as exc:
                    print(f"Error loading dataset {name} | {exc}")
                pbar.update(1)
    return datasets


def check_updated_dataset(original: Dataset, new: Dataset) -> None:  # noqa: F821
    assert original.metadata.name == new.metadata.name, "Datasets have different names"
    assert (
        original.metadata.indicator_metadata == new.metadata.indicator_metadata
    ), "Datasets have different indicator metadata"
    assert (
        original.data.shape[1] == new.data.shape[1]
    ), "Datasets have different number of columns"
    assert (
        original.data.index[0] == new.data.index[0]
    ), "Datasets have different start date"

    shortened_n = int(original.data.shape[0] * 0.9)
    shortened_original = original.data.head(shortened_n)
    shortened_new = new.data.head(shortened_n)
    assert (
        shortened_original.notna().sum().sum() == shortened_new.notna().sum().sum()
    ), "Datasets have different number of missing values"
    try:
        pd.testing.assert_series_equal(
            shortened_original.mean(), shortened_new.mean(), atol=0, rtol=0.05
        )
    except AssertionError as exc:
        raise AssertionError("Datasets have different means") from exc
    try:
        pd.testing.assert_series_equal(
            shortened_original.std(), shortened_new.std(), atol=0, rtol=0.05
        )
    except AssertionError as exc:
        raise AssertionError("Datasets have different standard deviations") from exc



import datetime as dt
import re
import tempfile
import zipfile
from io import BytesIO
from os import path
from urllib.error import URLError

import pandas as pd
import httpx
from pandas.tseries.offsets import MonthEnd, YearEnd

from econuy import load_dataset
from econuy.base import Dataset, DatasetMetadata
from econuy.retrieval import regional
from econuy.utils.operations import get_download_sources, get_name_from_function
from econuy.utils.extras import TRADE_METADATA, BOP_COLUMNS
from econuy.utils.retrieval import get_with_ssl_context


def _get_trade(dataset_name: str) -> Dataset:
    """Helper function. See any of the `trade_...()` functions."""
    sources = get_download_sources(dataset_name)
    meta = TRADE_METADATA[dataset_name]
    try:
        xls = pd.ExcelFile(sources["main"])
    except URLError as err:
        if "SSL: CERTIFICATE_VERIFY_FAILED" in str(err):
            r_bytes = get_with_ssl_context("bcu", sources["main"])
            xls = pd.ExcelFile(r_bytes)
    sheets = []
    start_col = meta["start_col"]
    for sheet in xls.sheet_names:
        raw = (
            pd.read_excel(xls, sheet_name=sheet, index_col=start_col, skiprows=7)
            .iloc[:, start_col:]
            .dropna(thresh=5)
            .T
        )
        raw.index = pd.to_datetime(raw.index, errors="coerce") + MonthEnd(0)
        proc = raw[raw.index.notnull()].dropna(thresh=5, axis=1)
        if dataset_name != "trade_imports_category_value":
            try:
                proc = proc.loc[:, meta["colnames"].keys()]
            except KeyError:
                proc.insert(7, "Venezuela", 0)
                proc = proc.loc[:, meta["colnames"].keys()]
            proc.columns = meta["colnames"].values()
        else:
            proc = proc.loc[:, ~(proc == "miles de dólares").any()]
            proc = proc.drop(columns=["DESTINO ECONÓMICO"])
            proc.columns = meta["new_colnames"]
        sheets.append(proc)
    output = pd.concat(sheets).sort_index()
    output = output.apply(pd.to_numeric, errors="coerce")
    if meta["unit"] == "Millions":
        output = output.div(1000)
    output = output.rename_axis(None)

    spanish_names = output.columns
    spanish_names = [{"es": x} for x in spanish_names]
    ids = [f"{dataset_name}_{i}" for i in range(output.shape[1])]
    output.columns = ids

    base_metadata = {
        "area": "External sector",
        "currency": meta["currency"],
        "inflation_adjustment": None,
        "unit": meta["unit"],
        "seasonal_adjustment": None,
        "frequency": "ME",
        "time_series_type": "Flow",
        "cumulative_periods": 1,
        "transformations": [],
    }
    metadata = DatasetMetadata.from_cast(
        dataset_name, base_metadata, output.columns, spanish_names
    )
    dataset = Dataset(dataset_name, output, metadata)

    return dataset


def trade_exports_sector_value() -> Dataset:
    """Get export values by product.

    Returns
    -------
    Export values by product : Dataset

    """
    name = get_name_from_function()
    return _get_trade(name)


def trade_exports_sector_volume() -> Dataset:
    """Get export volumes by product.

    Returns
    -------
    Export volumes by product : Dataset

    """
    name = get_name_from_function()
    return _get_trade(name)


def trade_exports_sector_price() -> Dataset:
    """Get export prices by product.

    Returns
    -------
    Export prices by product : Dataset

    """
    name = get_name_from_function()
    return _get_trade(name)


def trade_exports_destination_value() -> Dataset:
    """Get export values by destination.

    Returns
    -------
    Export values by destination : Dataset

    """
    name = get_name_from_function()
    return _get_trade(name)


def trade_exports_destination_volume() -> Dataset:
    """Get export volumes by destination.

    Returns
    -------
    Export volumes by destination : Dataset

    """
    name = get_name_from_function()
    return _get_trade(name)


def trade_exports_destination_price() -> Dataset:
    """Get export prices by destination.

    Returns
    -------
    Export prices by destination : Dataset

    """
    name = get_name_from_function()
    return _get_trade(name)


def trade_imports_category_value() -> Dataset:
    """Get import values by sector.

    Returns
    -------
    Import values by sector : Dataset

    """
    name = get_name_from_function()
    return _get_trade(name)


def trade_imports_category_volume() -> Dataset:
    """Get import volumes by sector.

    Returns
    -------
    Import volumes by sector : Dataset

    """
    name = get_name_from_function()
    return _get_trade(name)


def trade_imports_category_price() -> Dataset:
    """Get import prices by sector.

    Returns
    -------
    Import prices by sector : Dataset

    """
    name = get_name_from_function()
    return _get_trade(name)


def trade_imports_origin_value() -> Dataset:
    """Get import values by origin.

    Returns
    -------
    Import values by origin : Dataset

    """
    name = get_name_from_function()
    return _get_trade(name)


def trade_imports_origin_volume() -> Dataset:
    """Get import volumes by origin.

    Returns
    -------
    Import volumes by origin : Dataset

    """
    name = get_name_from_function()
    return _get_trade(name)


def trade_imports_origin_price() -> Dataset:
    """Get import prices by origin.

    Returns
    -------
    Import prices by origin : Dataset

    """
    name = get_name_from_function()
    return _get_trade(name)


def trade_balance(*args, **kwargs) -> Dataset:
    """
    Get net trade balance data by country/region.

    Returns
    -------
    Net trade balance value by region/country : Dataset

    """
    name = get_name_from_function()
    exports = (
        load_dataset("trade_exports_destination_value", *args, **kwargs)
        .to_named()
        .rename(columns={"Total exportaciones": "Total"})
    )
    imports = (
        load_dataset("trade_imports_origin_value", *args, **kwargs)
        .to_named()
        .rename(columns={"Total importaciones": "Total"})
    )
    output = exports - imports
    output = output.rename_axis(None)

    spanish_names = exports.columns
    spanish_names = [{"es": x} for x in spanish_names]
    ids = [f"{name}_{i}" for i in range(output.shape[1])]
    output.columns = ids

    base_metadata = {
        "area": "External sector",
        "currency": "USD",
        "inflation_adjustment": None,
        "unit": "Millions",
        "seasonal_adjustment": None,
        "frequency": "ME",
        "time_series_type": "Flow",
        "cumulative_periods": 1,
        "transformations": [],
    }
    metadata = DatasetMetadata.from_cast(
        name, base_metadata, output.columns, spanish_names
    )
    dataset = Dataset(name, output, metadata)

    return dataset


def terms_of_trade(*args, **kwargs) -> Dataset:
    """
    Get terms of trade.

    Returns
    -------
    Terms of trade (exports/imports) : Dataset

    """
    name = get_name_from_function()
    exports = (
        load_dataset("trade_exports_destination_price", *args, **kwargs)
        .to_named()
        .rename(columns={"Total exportaciones": "Total"})
    )
    imports = (
        load_dataset("trade_imports_origin_price", *args, **kwargs)
        .to_named()
        .rename(columns={"Total importaciones": "Total"})
    )

    output = exports / imports
    output = output.loc[:, ["Total"]]
    output = output.rename(columns={"Total": "Términos de intercambio"})
    output = output.rename_axis(None)

    spanish_names = exports.columns
    spanish_names = [{"es": x} for x in spanish_names]
    ids = [f"{name}_{i}" for i in range(output.shape[1])]
    output.columns = ids

    base_metadata = {
        "area": "External sector",
        "currency": "USD",
        "inflation_adjustment": None,
        "unit": "2005=100",
        "seasonal_adjustment": None,
        "frequency": "ME",
        "time_series_type": "Flow",
        "cumulative_periods": 1,
        "transformations": [],
    }
    metadata = DatasetMetadata.from_cast(
        name, base_metadata, output.columns, spanish_names
    )
    dataset = Dataset(name, output, metadata)
    dataset = dataset.rebase("2005-01-01", "2005-12-31", 100)
    dataset.metadata.update_dataset_metadata({"unit": "2005=100"})
    dataset.transformed = False

    return dataset


def _commodity_weights() -> pd.DataFrame:
    raw = pd.read_csv(
        "https://raw.githubusercontent.com/rxavier/econuy-extras/main/econuy_extras/manual_data/comtrade.csv"
    )

    table = raw.groupby(["RefYear", "CmdDesc"]).sum().reset_index()
    table = table.pivot(index="RefYear", columns="CmdDesc", values="PrimaryValue")
    table.fillna(0, inplace=True)

    beef = [
        "Bovine animals, live",
        "Bovine meat, fresh, chilled or frozen",
        "Bovine meat,frsh,chilled",
        "Edible offal of bovine animals, frozen",
        "Edible offal of bovine animals, fresh or chilled",
        "Edible offal of bovine animals, fresh/chilled",
        "Meat & offal (other than liver), of bovine animals, prepared/preserved, n.e.s.",
        "Meat and offal (other than liver), of bovine animals, prepared or preserv",
        "Meat of bovine animals, fresh or chilled",
        "Meat of bovine animals,fresh,chilled or frozen",
    ]
    table["Beef"] = table[beef].sum(axis=1, min_count=len(beef))
    table.drop(beef, axis=1, inplace=True)
    table.columns = [
        "Cebada",
        "Madera",
        "Oro",
        "Leche",
        "Pulpa de celulosa",
        "Arroz",
        "Soja",
        "Trigo",
        "Lana",
        "Carne bovina",
    ]
    output = table.div(table.sum(axis=1), axis=0)
    output.index = pd.to_datetime(output.index, format="%Y") + YearEnd(1)
    output = output.rolling(window=3, min_periods=3).mean().bfill()

    return output


def commodity_prices() -> Dataset:
    """Get commodity prices for Uruguay.

    Returns
    -------
    Commodity prices : Dataset
        Prices and price indexes of relevant commodities for Uruguay.

    """
    name = get_name_from_function()
    sources = get_download_sources(name)

    raw_beef = pd.read_excel(
        sources["beef"], header=4, index_col=0, thousands=".", usecols="A:D"
    ).dropna(how="all")

    raw_beef.columns = raw_beef.columns.str.strip()
    proc_beef = raw_beef["Ing. Prom./Ton."].to_frame()
    proc_beef.index = pd.date_range(
        start="2002-01-04", periods=len(proc_beef), freq="W-SAT"
    )
    beef = proc_beef.resample("ME").mean()

    milk_r = httpx.get(sources["milk1"])
    xls = re.findall(
        r"https://www.inale.org/wp-content/uploads/[0-9\/]+/Precios-exportacion-de-Europa.xls",
        milk_r.text,
        flags=re.IGNORECASE,
    )[0]
    raw_milk = pd.read_excel(
        xls,
        skiprows=13,
        nrows=dt.datetime.now().year - 2006,
    )
    raw_milk.dropna(how="all", axis=1, inplace=True)
    raw_milk.drop(["Promedio ", "Variación"], axis=1, inplace=True)
    raw_milk.columns = ["Año/Mes"] + list(range(1, 13))
    proc_milk = pd.melt(raw_milk, id_vars=["Año/Mes"])
    proc_milk["Año/Mes"] = pd.to_numeric(proc_milk["Año/Mes"], errors="coerce")
    proc_milk = proc_milk.dropna(subset=["Año/Mes"])
    proc_milk = proc_milk.sort_values(by=["Año/Mes", "variable"])
    proc_milk.index = pd.date_range(
        start="2007-01-31", periods=len(proc_milk), freq="ME"
    )
    proc_milk = proc_milk.iloc[:, 2].to_frame().divide(10).dropna()

    prev_milk = pd.read_excel(
        sources["milk2"],
        sheet_name="Raw Milk Prices",
        index_col=0,
        skiprows=6,
        usecols="A:AB",
        na_values=["c", 0],
    )
    prev_milk = (
        prev_milk[prev_milk.index.notna()]
        .dropna(axis=0, how="all")
        .mean(axis=1)
        .to_frame()
    )
    prev_milk = prev_milk.set_index(
        pd.date_range(start="1977-01-31", freq="ME", periods=len(prev_milk))
    )
    eurusd_r = httpx.get(
        "https://fx.sauder.ubc.ca/cgi/fxdata",
        params=f"b=USD&c=EUR&rd=&fd=1&fm=1&fy=2001&ld=31&lm=12&ly="
        f"{dt.datetime.now().year}&y=monthly&q=volume&f=html&o=",
    )
    eurusd = pd.read_html(BytesIO(eurusd_r.content))[0].drop("MMM YYYY", axis=1)
    eurusd.index = pd.date_range(start="2001-01-31", periods=len(eurusd), freq="ME")
    eurusd_milk = eurusd.reindex(prev_milk.index)
    prev_milk = prev_milk.divide(eurusd_milk.values).multiply(10)
    prev_milk = prev_milk.loc[prev_milk.index < min(proc_milk.index)]
    prev_milk.columns, proc_milk.columns = ["Price"], ["Price"]
    milk = pd.concat([prev_milk, proc_milk])

    raw_pulp_r = httpx.get(sources["pulp"].format(year=dt.date.today().year))
    temp_dir = tempfile.TemporaryDirectory()
    with zipfile.ZipFile(BytesIO(raw_pulp_r.content), "r") as f:
        f.extractall(path=temp_dir.name)
        path_temp = path.join(temp_dir.name, "monthly_values.csv")
        raw_pulp = pd.read_csv(path_temp, sep=";").dropna(how="any")
    proc_pulp = raw_pulp.copy().sort_index(ascending=False)
    proc_pulp.index = pd.date_range(
        start="1990-01-31", periods=len(proc_pulp), freq="ME"
    )
    proc_pulp = proc_pulp.drop(["Label", "Codes"], axis=1).astype(float)
    proc_pulp = proc_pulp.div(eurusd.reindex(proc_pulp.index).values)
    pulp = proc_pulp

    r_imf = httpx.get(sources["imf"])
    imf = re.findall("external-data.+ashx", r_imf.text)[0]
    imf = f"https://imf.org/-/media/Files/Research/CommodityPrices/Monthly/{imf}"
    raw_imf = pd.read_excel(imf).dropna(how="all", axis=1).dropna(how="all", axis=0)
    raw_imf.columns = raw_imf.iloc[0, :]
    proc_imf = raw_imf.iloc[3:, 1:]
    proc_imf.index = pd.date_range(start="1990-01-31", periods=len(proc_imf), freq="ME")
    rice = proc_imf[proc_imf.columns[proc_imf.columns.str.contains("Rice")]]
    wood = proc_imf[proc_imf.columns[proc_imf.columns.str.contains("Sawnwood")]]
    wood = wood.mean(axis=1).to_frame()
    wool = proc_imf[proc_imf.columns[proc_imf.columns.str.startswith("Wool")]]
    wool = wool.mean(axis=1).to_frame()
    barley = proc_imf[proc_imf.columns[proc_imf.columns.str.startswith("Barley")]]
    gold = proc_imf[proc_imf.columns[proc_imf.columns.str.startswith("Gold")]]
    soybean = proc_imf[
        proc_imf.columns[proc_imf.columns.str.startswith("Soybeans, U.S.")]
    ]
    wheat = proc_imf[proc_imf.columns[proc_imf.columns.str.startswith("Wheat")]]

    output = pd.concat(
        [beef, pulp, soybean, milk, rice, wood, wool, barley, gold, wheat], axis=1
    )
    output = output.reindex(beef.index).dropna(thresh=8)
    output.columns = [
        "Carne bovina",
        "Pulpa de celulosa",
        "Soja",
        "Leche",
        "Arroz",
        "Madera",
        "Lana",
        "Cebada",
        "Oro",
        "Trigo",
    ]
    output = output.apply(pd.to_numeric, errors="coerce")
    output = output.rename_axis(None)

    spanish_names = output.columns
    spanish_names = [{"es": x} for x in spanish_names]
    ids = [f"{name}_{i}" for i in range(output.shape[1])]
    output.columns = ids

    base_metadata = {
        "area": "External sector",
        "currency": "USD",
        "inflation_adjustment": None,
        "unit": "-",
        "seasonal_adjustment": None,
        "frequency": "ME",
        "time_series_type": "Flow",
        "cumulative_periods": 1,
        "transformations": [],
    }
    metadata = DatasetMetadata.from_cast(
        name, base_metadata, output.columns, spanish_names
    )
    units = [
        "USD per ton",
        "USD per ton",
        "USD per ton",
        "USD per ton",
        "USD per ton",
        "USD per m3",
        "US cent. per kg",
        "USD per ton",
        "USD per onza troy",
        "USD per ton",
    ]
    for indicator, unit in zip(ids, units):
        metadata.update_indicator_metadata_value(indicator, "unit", unit)

    dataset = Dataset(name, output, metadata)

    return dataset


def commodity_index(*args, **kwargs) -> Dataset:
    """Get export-weighted commodity price index for Uruguay.

    Returns
    -------
    Monthly export-weighted commodity index : Dataset
        Export-weighted average of commodity prices relevant to Uruguay.

    """
    name = get_name_from_function()
    prices = load_dataset("commodity_prices", *args, **kwargs).to_named()
    prices = prices.interpolate(method="linear", limit=1).dropna(how="any")
    prices = prices.pct_change(periods=1)
    weights = _commodity_weights()
    weights = weights[prices.columns]
    weights = weights.reindex(prices.index, method="ffill")

    output = pd.DataFrame(
        prices.values * weights.values, columns=prices.columns, index=prices.index
    )
    output = output.sum(axis=1).add(1).to_frame().cumprod().multiply(100)
    output.columns = ["Índice de precios de productos primarios"]
    output = output.rename_axis(None)

    spanish_names = output.columns
    spanish_names = [{"es": x} for x in spanish_names]
    ids = [f"{name}_{i}" for i in range(output.shape[1])]
    output.columns = ids

    base_metadata = {
        "area": "External sector",
        "currency": "USD",
        "inflation_adjustment": None,
        "unit": "2002-01=100",
        "seasonal_adjustment": None,
        "frequency": "ME",
        "time_series_type": "Flow",
        "cumulative_periods": 1,
        "transformations": [],
    }
    metadata = DatasetMetadata.from_cast(
        name, base_metadata, output.columns, spanish_names
    )
    dataset = Dataset(name, output, metadata)

    return dataset


def rxr() -> Dataset:
    """Get official (BCU) real exchange rates.

    Returns
    -------
    Monthly real exchange rates vs select countries/regions : Dataset
        Available: global, regional, extraregional, Argentina, Brazil, US.

    """
    name = get_name_from_function()
    sources = get_download_sources(name)
    try:
        raw = pd.read_excel(sources["main"], skiprows=8, usecols="B:N", index_col=0)
    except URLError as err:
        if "SSL: CERTIFICATE_VERIFY_FAILED" in str(err):
            r_bytes = get_with_ssl_context("bcu", sources["main"])
            raw = pd.read_excel(r_bytes, skiprows=9, usecols="B:N", index_col=0)
    output = raw.dropna(how="any")
    output.columns = [
        "Global",
        "Extrarregional",
        "Regional",
        "Argentina",
        "Brasil",
        "EE.UU.",
        "México",
        "Alemania",
        "España",
        "Reino Unido",
        "Italia",
        "China",
    ]
    output.index = pd.to_datetime(output.index) + MonthEnd(1)
    output = output.rename_axis(None)

    spanish_names = output.columns
    spanish_names = [{"es": x} for x in spanish_names]
    ids = [f"{name}_{i}" for i in range(output.shape[1])]
    output.columns = ids

    base_metadata = {
        "area": "External sector",
        "currency": "UYU/other",
        "inflation_adjustment": None,
        "unit": "2019=100",
        "seasonal_adjustment": None,
        "frequency": "ME",
        "time_series_type": "Flow",
        "cumulative_periods": 1,
        "transformations": [],
    }
    metadata = DatasetMetadata.from_cast(
        name, base_metadata, output.columns, spanish_names
    )
    dataset = Dataset(name, output, metadata)

    return dataset


def rxr_custom(*args, **kwargs) -> Dataset:
    """Get custom real exchange rates vis-à-vis the US, Argentina and Brazil.

    Returns
    -------
    Monthly real exchange rates vs select countries : Dataset
        Available: Argentina, Brazil, US.

    """
    name = get_name_from_function()

    ifs = regional._ifs(*args, **kwargs)
    uy_cpi = load_dataset("cpi", *args, **kwargs).to_named()
    uy_e = load_dataset("nxr_monthly", *args, **kwargs).to_named().iloc[:, [1]]
    proc = pd.concat([ifs, uy_cpi, uy_e], axis=1)
    proc = proc.interpolate(method="linear", limit_area="inside")
    proc = proc.dropna(how="all")
    proc.columns = ["AR_E_O", "AR_E_U", "BR_E", "AR_P", "BR_P", "US_P", "UY_P", "UY_E"]

    output = pd.DataFrame()
    output["UY_E_P"] = proc["UY_E"] / proc["UY_P"]
    output["Uruguay-Argentina oficial"] = (
        output["UY_E_P"] / proc["AR_E_O"] * proc["AR_P"]
    )
    output["Uruguay-Argentina informal"] = (
        output["UY_E_P"] / proc["AR_E_U"] * proc["AR_P"]
    )
    output["Uruguay-Brasil"] = output["UY_E_P"] / proc["BR_E"] * proc["BR_P"]
    output["Uruguay-EE.UU."] = output["UY_E_P"] * proc["US_P"]
    output = output.drop("UY_E_P", axis=1)
    output = output.loc[output.index >= "1979-12-01"]
    output = output.rename_axis(None)
    output = output.dropna(how="all")

    spanish_names = output.columns
    spanish_names = [{"es": x} for x in spanish_names]
    ids = [f"{name}_{i}" for i in range(output.shape[1])]
    output.columns = ids

    base_metadata = {
        "area": "External sector",
        "currency": "UYU/other",
        "inflation_adjustment": None,
        "unit": "2019=100",
        "seasonal_adjustment": None,
        "frequency": "ME",
        "time_series_type": "Flow",
        "cumulative_periods": 1,
        "transformations": [],
    }
    metadata = DatasetMetadata.from_cast(
        name, base_metadata, output.columns, spanish_names
    )
    for indicator, currency in zip(ids, ["UYU/ARS", "UYU/ARS", "UYU/BRL", "UYU/USD"]):
        metadata.update_indicator_metadata_value(indicator, "currency", currency)
    dataset = Dataset(name, output, metadata).rebase("2010-01-01", "2010-12-31", 100)
    dataset.metadata.update_dataset_metadata({"unit": "2010=100"})
    dataset.transformed = False

    return dataset


def balance_of_payments() -> Dataset:
    """Get balance of payments.

    Returns
    -------
    Quarterly balance of payments : Dataset

    """
    name = get_name_from_function()
    sources = get_download_sources(name)
    try:
        raw = (
            pd.read_excel(
                sources["main"], skiprows=7, index_col=0, sheet_name="Cuadro Nº 1"
            )
            .dropna(how="all")
            .T
        )
    except URLError as err:
        if "SSL: CERTIFICATE_VERIFY_FAILED" in str(err):
            r_bytes = get_with_ssl_context("bcu", sources["main"])
            raw = (
                pd.read_excel(
                    r_bytes,
                    skiprows=7,
                    index_col=0,
                    sheet_name="Cuadro Nº 1",
                )
                .dropna(how="all")
                .T
            )
    output = raw.iloc[:, 2:]
    output.index = pd.date_range(start="2012-03-31", freq="QE-DEC", periods=len(output))
    pattern = r"\(1\)|\(2\)|\(3\)|\(4\)|\(5\)"
    output.columns = [re.sub(pattern, "", x).strip() for x in output.columns]
    output = output.drop(
        [
            "Por Sector Institucional",
            "Por Categoría Funcional",
            "Por Instrumento y Sector Institucional",
        ],
        axis=1,
    )
    output.columns = BOP_COLUMNS
    output = output.apply(pd.to_numeric, errors="coerce")

    spanish_names = output.columns
    spanish_names = [{"es": x} for x in spanish_names]
    ids = [f"{name}_{i}" for i in range(output.shape[1])]
    output.columns = ids

    base_metadata = {
        "area": "External sector",
        "currency": "USD",
        "inflation_adjustment": None,
        "unit": "Millions",
        "seasonal_adjustment": None,
        "frequency": "ME",
        "time_series_type": "Flow",
        "cumulative_periods": 1,
        "transformations": [],
    }
    metadata = DatasetMetadata.from_cast(
        name, base_metadata, output.columns, spanish_names
    )
    dataset = Dataset(name, output, metadata)

    return dataset


def balance_of_payments_summary(*args, **kwargs) -> Dataset:
    """Get a balance of payments summary and capital flows calculations.

    Returns
    -------
    Quarterly balance of payments summary : Dataset

    """
    name = get_name_from_function()
    bop = load_dataset("balance_of_payments", *args, **kwargs).to_named()

    output = pd.DataFrame(index=bop.index)
    output["Cuenta corriente"] = bop["Cuenta Corriente"]
    output["Balance de bienes y servicios"] = bop["Bienes y Servicios"]
    output["Balance de bienes"] = bop["Bienes"]
    output["Exportaciones de bienes"] = bop["Bienes - Crédito"]
    output["Importaciones de bienes"] = bop["Bienes - Débito"]
    output["Balance de servicios"] = bop["Servicios"]
    output["Exportaciones de servicios"] = bop["Servicios - Crédito"]
    output["Importaciones de servicios"] = bop["Servicios - Débito"]
    output["Ingreso primario"] = bop["Ingreso Primario"]
    output["Ingreso secundario"] = bop["Ingreso Secundario"]
    output["Cuenta capital"] = bop["Cuenta Capital"]
    output["Crédito en cuenta capital"] = bop["Cuenta Capital - Crédito"]
    output["Débito en cuenta capital"] = bop["Cuenta Capital - Débito"]
    output["Cuenta financiera"] = bop["Cuenta Financiera"]
    output["Balance de inversión directa"] = bop["Inversión directa"]
    output["Inversión directa en el exterior"] = bop[
        "Inversión directa - Adquisición neta de activos financieros"
    ]
    output["Inversión directa en Uruguay"] = bop[
        "Inversión directa - Pasivos netos incurridos"
    ]
    output["Balance de inversión de cartera"] = bop["Inversión de cartera"]
    output["Inversión de cartera en el exterior"] = bop[
        "Inversión de cartera - Adquisición neta de activos fin"
    ]
    output["Inversión de cartera en Uruguay"] = bop[
        "Inversión de cartera - Pasivos netos incurridos"
    ]
    output["Saldo de derivados financieros"] = bop[
        "Derivados financieros - distintos de reservas"
    ]
    output["Balance de otra inversión"] = bop["Otra inversión"]
    output["Otra inversión en el exterior"] = bop[
        "Otra inversión - Adquisición neta de activos financieros"
    ]
    output["Otra inversión en Uruguay"] = bop[
        "Otra inversión - Pasivos netos incurridos"
    ]
    output["Variación de activos de reserva"] = bop["Activos de Reserva BCU"]
    output["Errores y omisiones"] = bop["Errores y Omisiones"]
    output["Flujos brutos de capital"] = (
        output["Inversión directa en Uruguay"]
        + output["Inversión de cartera en Uruguay"]
        + output["Otra inversión en Uruguay"]
        + output["Crédito en cuenta capital"]
    )
    output["Flujos netos de capital"] = (
        -output["Balance de inversión directa"]
        - output["Balance de inversión de cartera"]
        - output["Balance de otra inversión"]
        - output["Saldo de derivados financieros"]
        + output["Cuenta capital"]
    )

    spanish_names = output.columns
    spanish_names = [{"es": x} for x in spanish_names]
    ids = [f"{name}_{i}" for i in range(output.shape[1])]
    output.columns = ids

    base_metadata = {
        "area": "External sector",
        "currency": "USD",
        "inflation_adjustment": None,
        "unit": "Millions",
        "seasonal_adjustment": None,
        "frequency": "ME",
        "time_series_type": "Flow",
        "cumulative_periods": 1,
        "transformations": [],
    }
    metadata = DatasetMetadata.from_cast(
        name, base_metadata, output.columns, spanish_names
    )
    dataset = Dataset(name, output, metadata)

    return dataset


def international_reserves() -> Dataset:
    """Get international reserves data.

    Returns
    -------
    Daily international reserves : Dataset

    """
    name = get_name_from_function()
    sources = get_download_sources(name)
    try:
        raw = pd.read_excel(
            sources["main"], usecols="D:J", index_col=0, skiprows=5, na_values="n/d"
        )
    except URLError as err:
        if "SSL: CERTIFICATE_VERIFY_FAILED" in str(err):
            r_bytes = get_with_ssl_context("bcu", sources["main"])
            raw = pd.read_excel(
                r_bytes,
                usecols="D:J",
                index_col=0,
                skiprows=5,
                na_values="n/d",
            )
    proc = raw.dropna(thresh=1)
    output = proc[proc.index.notnull()]
    output.columns = [
        "Activos de reserva",
        "Otros activos externos de corto plazo",
        "Obligaciones en ME con el sector público",
        "Obligaciones en ME con el sector financiero",
        "Activos de reserva sin sector público y financiero",
        "Posición en ME del BCU",
    ]
    output = output.apply(pd.to_numeric, errors="coerce")
    output = output.loc[~output.index.duplicated(keep="first")]
    output.index = pd.to_datetime(output.index)
    output = output.rename_axis(None)

    spanish_names = output.columns
    spanish_names = [{"es": x} for x in spanish_names]
    ids = [f"{name}_{i}" for i in range(output.shape[1])]
    output.columns = ids

    base_metadata = {
        "area": "External sector",
        "currency": "USD",
        "inflation_adjustment": None,
        "unit": "Millions",
        "seasonal_adjustment": None,
        "frequency": "ME",
        "time_series_type": "Stock",
        "cumulative_periods": 1,
        "transformations": [],
    }
    metadata = DatasetMetadata.from_cast(
        name, base_metadata, output.columns, spanish_names
    )
    dataset = Dataset(name, output, metadata)

    return dataset


# @retry(
#     retry_on_exceptions=(HTTPError, URLError),
#     max_calls_total=10,
#     retry_window_after_first_call_in_seconds=90,
# )
# def international_reserves_changes(
#     pipeline: Optional[Pipeline] = None, previous_data: pd.DataFrame = pd.DataFrame()
# ) -> pd.DataFrame:
#     """Get international reserves changes data.

#     Parameters
#     ----------
#     pipeline : econuy.core.Pipeline or None, default None
#         An instance of the econuy Pipeline class.
#     previous_data : pd.DataFrame
#         A DataFrame representing this dataset used to extract last
#         available dates.

#     Returns
#     -------
#     Monthly international reserves changes : pd.DataFrame

#     """
#     name = get_name_from_function()
#     sources = get_download_sources(name)

#     if pipeline is None:
#         pipeline = Pipeline()
#     if previous_data.empty:
#         first_year = 2013
#     else:
#         first_year = previous_data.index[-1].year

#     mapping = dict(
#         zip(
#             [
#                 "Ene",
#                 "Feb",
#                 "Mar",
#                 "Abr",
#                 "May",
#                 "Jun",
#                 "Jul",
#                 "Ago",
#                 "Set",
#                 "Oct",
#                 "Nov",
#                 "Dic",
#             ],
#             [str(x).zfill(2) for x in range(1, 13)],
#         )
#     )
#     inverse_mapping = {v: k for k, v in mapping.items()}
#     mapping.update({"Sep": "09"})

#     current_year = dt.date.today().year
#     months = []
#     for year in range(first_year, current_year + 1):
#         if year < current_year:
#             filename = f"dic{year}.xls"
#         else:
#             current_month = inverse_mapping[str(dt.date.today().month).zfill(2)]
#             last_month = inverse_mapping[
#                 str((dt.date.today() + relativedelta(months=-1)).month).zfill(2)
#             ]
#             certs_path = Path(get_project_root(), "utils", "files", "bcu_certs.pem")
#             filename = f"{current_month}{year}.xls"
#             r = httpx.get(f"{sources['main']}{filename}", verify=certs_path)
#             if r.status_code == 404:
#                 filename = f"{last_month}{year}.xls"
#         try:
#             data = pd.read_excel(
#                 f"{sources['main']}{filename}",
#                 skiprows=2,
#                 sheet_name="ACTIVOS DE RESERVA",
#             )
#         except URLError as err:
#             if "SSL: CERTIFICATE_VERIFY_FAILED" in str(err):
#                 certs_path = Path(get_project_root(), "utils", "files", "bcu_certs.pem")
#                 r = httpx.get(f"{sources['main']}{filename}", verify=certs_path)
#                 data = pd.read_excel(
#                     r.content,
#                     skiprows=2,
#                     sheet_name="ACTIVOS DE RESERVA",
#                 )
#         data = data.dropna(how="all").dropna(how="all", axis=1).set_index("CONCEPTOS")
#         if data.columns[0] == "Mes":
#             data.columns = data.iloc[0, :]
#         data = (
#             data.iloc[1:]
#             .T.reset_index(names="date")
#             .loc[
#                 lambda x: ~x["date"]
#                 .astype(str)
#                 .str.contains("Unnamed|Trimestre|Año|I", regex=True, case=True),
#                 lambda x: x.columns.notna(),
#             ]
#         )
#         data["date"] = data["date"].replace("Mes\n", "", regex=True).str.strip()
#         data = data.loc[data["date"].notna()]

#         index = pd.Series(data["date"]).str.split("-", expand=True).replace(mapping)
#         index = pd.to_datetime(
#             index.iloc[:, 0] + "-" + index.iloc[:, 1], format="%m-%Y", errors="coerce"
#         ) + pd.tseries.offsets.MonthEnd(1)
#         if year == 2019:
#             index = index.fillna(dt.datetime(year, 1, 31))
#         elif year == 2013:
#             index = index.fillna(dt.datetime(year, 12, 31))
#         data["date"] = index
#         data.columns = ["date"] + RESERVES_COLUMNS
#         months.append(data)
#     reserves = (
#         pd.concat(months, sort=False, ignore_index=True)
#         .drop_duplicates(subset="date")
#         .dropna(subset="date")
#         .set_index("date")
#         .sort_index()
#         .rename_axis(None)
#     )
#     metadata._set(
#         reserves,
#         area="Sector externo",
#         currency="USD",
#         inf_adj="No",
#         unit="Millones",
#         seas_adj="NSA",
#         ts_type="Flujo",
#         cumperiods=1,
#     )
#     reserves.columns = reserves.columns.set_levels(["-"], level=2)

#     return reserves
